
MODULE MultiMigrate;

        (********************************************************)
        (*                                                      *)
        (*         Migrating Weasel to a multidomain            *)
        (*                      configuration                   *)
        (*                                                      *)
        (*  Programmer:         P. Moylan                       *)
        (*  Last edited:        2 March 2024                    *)
        (*  Status:             OK                              *)
        (*                                                      *)
        (* I am deliberately not supporting remote operations   *)
        (* in this program, because there's too much risk of    *)
        (* accidents. This program must be run on the same      *)
        (* machine where Weasel lives.                          *)
        (*                                                      *)
        (********************************************************)


IMPORT Strings;

FROM SYSTEM IMPORT ADR;

FROM Names IMPORT
    (* type *)  FilenameString, DomainName, UserName;

FROM MiscFuncs IMPORT
    (* type *)  CharArrayPointer, LocArrayPointer,
    (* proc *)  StringMatch;

FROM Inet2Misc IMPORT
    (* proc *)  NameIsNumeric, StringToIP, IPToString;

FROM INIData IMPORT
    (* type *)  HINI, StringReadState,
    (* proc *)  ChooseDefaultINI, OpenINIFile, INIValid, CloseINIFile,
                GetStringList, NextString, CloseStringList,
                ItemSize, INIGet, INIGetString, INIGetTrusted, INIPut, INIPutBinary,
                INIDeleteApp, INIDeleteKey;

FROM FileOps IMPORT
    (* type *)  ChanId,
    (* proc *)  OpenNewFile, CloseFile, Exists, MoveDirectory, CreateDir;

FROM Sockets IMPORT
    (* const*)  AF_INET, SOCK_DGRAM, AF_UNSPEC, IFMIB_ENTRIES, NotASocket,
    (* type *)  Socket,
    (* proc *)  socket, ioctl, soclose;

FROM NetIF IMPORT
    (* const*)  IFNAMSIZ,
    (* type *)  ifconf, ifreq;

FROM ioctl IMPORT
    (* const*)  SIOCGIFCONF;

FROM LowLevel IMPORT
    (* proc *)  AddOffset;

FROM Storage IMPORT
    (* proc *)  ALLOCATE, DEALLOCATE;

FROM STextIO IMPORT
    (* proc *)  WriteString, WriteLn, ReadString;

(************************************************************************)

CONST
    Nul = CHR(0);

VAR
    MasterINIFile, MailRoot: FilenameString;
    UseTNI: BOOLEAN;
    SYSapp: ARRAY [0..7] OF CHAR;

(************************************************************************)

PROCEDURE MoveOneUser (name: UserName;  INIFile1, INIFile2: FilenameString);

    (* Removes 'name' from INIFile1 and puts it into INIFile2.  We      *)
    (* don't move any files, that's the caller's job.                   *)

    TYPE ListOfKeys = POINTER TO RECORD
                                     next: ListOfKeys;
                                     key: UserName;
                                     size: CARDINAL;
                                     val: LocArrayPointer;
                                 END (*RECORD*);

    VAR hini1, hini2: HINI;
        keylist, listelt: ListOfKeys;
        state: StringReadState;
        key: UserName;

    BEGIN
        (* We start by reading all the user data into memory - after    *)
        (* which we'll add it to the second INI file.                   *)

        keylist := NIL;
        hini1 := OpenINIFile (INIFile1);
        IF INIValid (hini1) THEN
            key := "";
            GetStringList (hini1, name, key, state);
            REPEAT
                NextString (state, key);
                IF key[0] <> Nul THEN
                    NEW (listelt);
                    listelt^.key := key;
                    IF NOT ItemSize (hini1, name, key, listelt^.size) THEN
                        listelt^.size := 0;
                    END (*IF*);
                    IF listelt^.size = 0 THEN
                        listelt^.val := NIL;
                    ELSE
                        ALLOCATE (listelt^.val, listelt^.size);
                        IF NOT INIGetTrusted(hini1, name, key, listelt^.val^,
                                                        listelt^.size) THEN
                            listelt^.size := 0;
                        END (*IF*);
                    END (*IF*);
                    listelt^.next := keylist;
                    keylist := listelt;
                END (*IF*);
            UNTIL key[0] = Nul;
            CloseStringList (state);
            CloseINIFile (hini1);
        END (*IF*);

        (* Now open the second INI file and write the data to it. *)

        hini2 := OpenINIFile (INIFile2);
        IF INIValid (hini2) THEN
            WHILE keylist <> NIL DO
                listelt := keylist;
                keylist := listelt^.next;
                IF listelt^.size = 0 THEN
                    INIPutBinary (hini2, name, listelt^.key, key, 0);
                ELSE
                    INIPutBinary (hini2, name, listelt^.key, listelt^.val^, listelt^.size);
                    DEALLOCATE (listelt^.val, listelt^.size);
                END (*IF*);
                DISPOSE (listelt);
            END (*WHILE*);
            CloseINIFile (hini2);
        END (*IF*);

        (* Now that the copy is complete, remove the user from  *)
        (* the first INI file.                                  *)

        hini1 := OpenINIFile (INIFile1);
        IF INIValid (hini1) THEN
            INIDeleteApp (hini1, name);
            CloseINIFile (hini1);
        END (*IF*);

    END MoveOneUser;

(************************************************************************)

PROCEDURE MoveAllUsers (domain2: DomainName);

    (* Removes all users from the master INI file and puts them into    *)
    (* domain2.  This procedure also moves the mail files, the local    *)
    (* host names, and the IMAP flag.                                   *)

    TYPE ListOfHosts = POINTER TO RECORD
                                      next: ListOfHosts;
                                      host: DomainName;
                                  END (*RECORD*);

    VAR INIFile1, INIFile2: FilenameString;
        hostlist: ListOfHosts;

    (********************************************************************)

    PROCEDURE IsDuplicate (val: CARDINAL): BOOLEAN;

        (* Returns TRUE iff val duplicates an entry in hostlist. *)

        VAR p: ListOfHosts;

        BEGIN
            p := hostlist;
            WHILE p <> NIL DO
                IF NameIsNumeric(p^.host) THEN
                    IF StringToIP(p^.host) = val THEN
                        RETURN TRUE;
                    END (*IF*);
                END (*IF*);
                p := p^.next;
            END (*WHILE*);

            RETURN FALSE;

        END IsDuplicate;

    (********************************************************************)

    PROCEDURE AddAddress (val: CARDINAL);

        (* Adds a new entry to the hostlist list. *)

        VAR item: ListOfHosts;

        BEGIN
            IF NOT IsDuplicate (val) THEN
                NEW (item);
                IPToString (val, FALSE, item^.host);
                item^.next := hostlist;
                hostlist := item;
            END (*IF*);
        END AddAddress;

    (********************************************************************)

    PROCEDURE AddLocalAddresses;

        TYPE ifreqPtr = POINTER TO ifreq;

        VAR s: Socket;  i, total: CARDINAL;
            ifc: ifconf;
            ifr: ifreqPtr;
            buf: ARRAY [0..IFMIB_ENTRIES*SIZE(ifreq)-1] OF CHAR;

        BEGIN
            (* Use an ioctl call to read all the ifreq records      *)
            (* into buffer "buf", and set "total" to the number of  *)
            (* such records.                                        *)

            s := socket(AF_INET, SOCK_DGRAM, AF_UNSPEC);
            IF s = NotASocket THEN
                total := 0;
            ELSE
                ifc.ifc_len := SIZE(buf);
                ifc.ifcu_buf := ADR(buf);
                IF ioctl (s, SIOCGIFCONF, ifc, SIZE(ifc)) < 0 THEN
                    total := 0;
                ELSE
                    total := ifc.ifc_len DIV SIZE(ifreq);
                END (*IF*);
                soclose(s);
            END (*IF*);

            (* Now work out which of the interface records are for  *)
            (* active interfaces, and put those addresses into the  *)
            (* listbox (unless they are duplicates).                *)

            IF total > 0 THEN

                ifr := ADR(buf);
                FOR i := 0 TO total-1 DO

                    IF ifr^.ifru_addr.family = AF_INET THEN
                        AddAddress (ifr^.ifru_addr.in_addr.addr);
                    END (*IF*);
                    ifr := AddOffset (ifr, SIZE(ifreq));

                END (*FOR*);

            END (* IF total > 0 *);

        END AddLocalAddresses;

    (********************************************************************)

    PROCEDURE CollectLocalHostNames (hini: HINI);

        (* Reads the local host names from the INI file (assumed to be  *)
        (* already open) and adds them to the hostlist list.            *)

        VAR hostlistelt: ListOfHosts;
            hostname: DomainName;
            state: StringReadState;
            key: ARRAY [0..7] OF CHAR;

        BEGIN
            key := "Local";
            GetStringList (hini, SYSapp, key, state);
            REPEAT
                NextString (state, hostname);
                IF hostname[0] <> Nul THEN
                    NEW (hostlistelt);
                    hostlistelt^.host := hostname;
                    hostlistelt^.next := hostlist;
                    hostlist := hostlistelt;
                END (*IF*);
            UNTIL hostname[0] = Nul;
            CloseStringList (state);
        END CollectLocalHostNames;

    (********************************************************************)

    PROCEDURE MoveLocalList;

        (* Takes the 'hostlist' list and merges those entries into the  *)
        (* local list of the target INI file.                           *)

        VAR j, k, BufferSize: CARDINAL;
            hini: HINI;
            p: ListOfHosts;
            bufptr: CharArrayPointer;

        BEGIN
            (* We already have a list from the source INI file.  Now    *)
            (* add names, if any, from the target INI file.             *)

            hini := OpenINIFile (INIFile2);
            IF INIValid (hini) THEN
                CollectLocalHostNames (hini);
                CloseINIFile (hini);
            END (*IF*);

            (* Add all the local addresses like loopback as well. *)

            AddLocalAddresses;

            (* Now we want to write all those names back to the target  *)
            (* INI file, as a string of strings.  Start by counting how *)
            (* much space we need.                                      *)

            BufferSize := 0;
            p := hostlist;
            WHILE p <> NIL DO
                INC (BufferSize, LENGTH(p^.host)+1);
                p := p^.next;
            END (*WHILE*);

            (* Create the string buffer. *)

            IF BufferSize = 0 THEN
                bufptr := NIL;
            ELSE
                INC (BufferSize);
                ALLOCATE (bufptr, BufferSize);
            END (*IF*);

            (* Store all the names into the buffer. *)

            j := 0;
            WHILE hostlist <> NIL DO
                p := hostlist;
                hostlist := p^.next;
                k := 0;
                REPEAT
                    bufptr^[j] := p^.host[k];
                    INC (k);  INC (j);
                UNTIL (p^.host[k] = Nul) OR (k = SIZE(DomainName));
                bufptr^[j] := Nul;  INC (j);
                DISPOSE (p);
            END (*WHILE*);

            (* Write the buffer to the target INI file. *)

            hini := OpenINIFile (INIFile2);
            IF INIValid (hini) THEN
                IF BufferSize = 0 THEN
                    INIPutBinary (hini, SYSapp, "Local", j, 0);
                ELSE
                    bufptr^[j] := Nul;
                    INIPutBinary (hini, SYSapp, "Local", bufptr^, BufferSize);
                    DEALLOCATE (bufptr, BufferSize);
                END (*IF*);
                CloseINIFile (hini);
            END (*IF*);

            (* Remove the original list from the source INI file. *)

            hini := OpenINIFile (INIFile1);
            IF INIValid (hini) THEN
                INIDeleteKey (hini, SYSapp, "Local");
                CloseINIFile (hini);
            END (*IF*);

        END MoveLocalList;

    (********************************************************************)

    TYPE
        ListOfUsers = POINTER TO RECORD
                                     next: ListOfUsers;
                                     user: UserName;
                                 END (*RECORD*);

    VAR SrcRoot, DstRoot, SrcDir, DstDir: FilenameString;
        userlist, listelt: ListOfUsers;
        state: StringReadState;
        current: UserName;
        cid: ChanId;
        hini: HINI;
        key : ARRAY [0..7] OF CHAR;
        L: CARDINAL;
        IMAPFlag, IMFlagExists: BOOLEAN;

    BEGIN
        (* Set the source directory and source INI file. *)

        SrcRoot := MailRoot;
        INIFile1 := MasterINIFile;

        (* Work out the destination directory and INI file. *)

        DstRoot := MailRoot;
        Strings.Append (domain2, DstRoot);
        Strings.Append ('\', DstRoot);
        INIFile2 := DstRoot;
        IF UseTNI THEN
            Strings.Append ("DOMAIN.TNI", INIFile2);
        ELSE
            Strings.Append ("DOMAIN.INI", INIFile2);
        END (*IF*);

        (* Make sure that the destination INI file exists.      *)

        CreateDir (DstRoot);
        IF NOT Exists (INIFile2) THEN
            cid := OpenNewFile (INIFile2, FALSE);
            CloseFile (cid);
        END (*IF*);

        (* Make a linear list of all users, and only then go back and   *)
        (* move user data.  Separating this into two steps makes the    *)
        (* code easier to read and check.                               *)

        userlist := NIL;
        hostlist := NIL;
        IMFlagExists := FALSE;
        hini := OpenINIFile (INIFile1);
        IF INIValid (hini) THEN

            (* Take the IMAPForNewUsers flag from the source file. *)

            IMFlagExists := INIGet (hini, SYSapp, "IMAPForNewUsers", IMAPFlag);
            IF IMFlagExists THEN
                INIDeleteKey (hini, SYSapp, "IMAPForNewUsers");
            END (*IF*);

            key := "";
            GetStringList (hini, key, key, state);
            REPEAT
                NextString (state, current);
                IF (current[0] <> Nul) AND NOT Strings.Equal(current, '$SYS') THEN
                    NEW (listelt);
                    listelt^.user := current;
                    listelt^.next := userlist;
                    userlist := listelt;
                END (*IF*);
            UNTIL current[0] = Nul;
            CloseStringList (state);

            (* While INIFile1 is still open, set a couple of other      *)
            (* details.                                                 *)

            INIPut (hini, SYSapp, 'MultiDomainEnabled', TRUE);
            L := Strings.Length (domain2);
            domain2[L] := Nul;  INC(L);
            domain2[L] := Nul;  INC(L);
            INIPutBinary (hini, SYSapp, 'Domains', domain2, L);
            CollectLocalHostNames(hini);
            CloseINIFile (hini);
        END (*IF*);

        (* Now that we have a list of all users, go back and move       *)
        (* them one at a time.  Remark: the user '$ALIAS' also gets     *)
        (* moved, thereby dealing with the aliases.                     *)

        WHILE userlist <> NIL DO
            listelt := userlist;
            current := listelt^.user;
            userlist := listelt^.next;
            DISPOSE (listelt);
            MoveOneUser (current, INIFile1, INIFile2);
            IF current[0] <> '$' THEN
                SrcDir := SrcRoot;
                Strings.Append (current, SrcDir);
                DstDir := DstRoot;
                Strings.Append (current, DstDir);
                IF NOT MoveDirectory (SrcDir, DstDir) THEN
                    WriteString ("ERROR: Failed to move the mail files.");
                END (*IF*);
            END (*IF*);
        END (*WHILE*);

        (* Put the IMAPForNewUsers flag into the destination file,      *)
        (* unless it already has such a flag.  Also set the multidomain *)
        (* mode.                                                        *)

        hini := OpenINIFile (INIFile2);
        IF INIValid (hini) THEN
            IF NOT INIGet (hini, SYSapp, "IMAPForNewUsers", IMAPFlag) THEN
                IF IMFlagExists THEN
                    INIPut (hini, SYSapp, "IMAPForNewUsers", IMAPFlag);
                ELSE
                    INIDeleteKey (hini, SYSapp, "IMAPForNewUsers");
                END (*IF*);
            END (*IF*);
            CloseINIFile (hini);
        END (*IF*);

        (* Finally, move the local host names over. *)

        MoveLocalList;

    END MoveAllUsers;

(************************************************************************)

PROCEDURE DoMigration;

    VAR multi, success: BOOLEAN;
        hini: HINI;
        newdomain: DomainName;

    BEGIN
        IF NOT ChooseDefaultINI ("Weasel", UseTNI) THEN
            UseTNI := FALSE;
        END (*IF*);
        IF UseTNI THEN
            MasterINIFile := "Weasel.TNI";
        ELSE
            MasterINIFile := "Weasel.INI";
        END (*IF*);
        success := TRUE;  multi := FALSE;
        IF NOT Exists (MasterINIFile) THEN
            WriteString ("ERROR: Cannot open the INI file");
            WriteLn;
            success := FALSE;
        END (*IF*);
        IF success THEN
            hini := OpenINIFile (MasterINIFile);
            IF INIValid (hini) THEN
                multi := INIGet (hini, SYSapp, 'MultiDomainEnabled', multi) AND multi;
                IF NOT INIGetString (hini, SYSapp, 'MailRoot', MailRoot) THEN
                    WriteString ("ERROR: MailRoot is not defined");
                    WriteLn;
                    success := FALSE;
                END (*IF*);
                CloseINIFile (hini);
            END (*IF*);
        END (*IF*);
        IF success THEN
            IF multi THEN
                WriteString ("Nothing to do.  You already have a multidomain configuration.");
                WriteLn;
                RETURN;
            END (*IF*);
            WriteString ("Enter your domain name > ");
            ReadString (newdomain);
            MoveAllUsers (newdomain);
            WriteString ("Finished. You may now run Weasel again.");
            WriteLn;
        END (*IF*);
    END DoMigration;

(************************************************************************)

BEGIN
    WriteString ("Program to convert Weasel to a multidomain configuration.");
    WriteLn;
    WriteString ("This must be executed from the mail Weasel directory.");
    WriteLn;
    WriteString ("Weasel.exe should not be running.");
    WriteLn;
    SYSapp := "$SYS";
    DoMigration;
END MultiMigrate.

